/*
 *	Simple HTML rendering engine
 *
 *	Currently we render compact html without forms support
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "parser.h"

struct render_state *render;		/* Current renderer */

void render_reset(struct render_state *r)
{
	render = r;
	render->rindent = 0;
	render->rbody = 0;
	render->rhead = 0;
	render->rfont = 0;
	render->rlayout = 0;
	render->rsize = 0;
	render->rlist = 0;
	free_anchors();
	renderer_reset();
}
	
static void render_indent(void)
{
	render->rindent++;
}

static void render_unindent(void)
{
	render->rindent--;
}

static void no_op(struct token_stack *t) { ; }

static void doc_begin(struct token_stack *t)
{
	renderer_begin(t);
}

static void doc_end(struct token_stack *t)
{
	renderer_complete(t);
}

static void a_begin(struct token_stack *t)
{
	char *href = tag_attr(t, "href");
	char *name = tag_attr(t, "name");

	token_peek_pop("A", NULL);
	
	t->rval = render->rfont & RF_LINK;

	if(href!=NULL)
	{
		render->rfont |= RF_LINK;
		link_area_begin(t, L_URLGOTO, href);
	}

	if(name!=NULL)
	{
		/* Add tag to the navigation */
		render_anchor(name);
	}
}

static void a_end(struct token_stack *t)
{
	if(tag_attr(t, "href")!=NULL)
		link_area_end(t, L_URLGOTO);
	render->rfont &= ~RF_LINK;
	render->rfont |= t->rval;
}

static void b_begin(struct token_stack *t)
{
	t->rval = render->rfont & RF_BOLD;
	render->rfont |= RF_BOLD;
}

static void b_end(struct token_stack *t)
{
	render->rfont &= ~RF_BOLD;
	render->rfont |= t->rval;
}

static void base_begin(struct token_stack *t)
{
	char *base = tag_attr(t, "href");
	if(base)
		set_base_url(base);
}

static void bq_begin(struct token_stack *t)
{
	t->rval = render->rfont & RF_BQUOTE;
	render->rfont |= RF_BQUOTE;
}

static void bq_end(struct token_stack *t)
{
	render->rfont &= ~RF_BQUOTE;
	render->rfont |= t->rval;
}

static void body_begin(struct token_stack *t)
{
	render->rbody = 1;
}

static void body_end(struct token_stack *t)
{
	render->rbody = 0;
}

static void br_begin(struct token_stack *t)
{
	/* FIXME: left right center - but do we care ?? */
	render_newline();
}

static void centre_begin(struct token_stack *t)
{
	render_newline_p();
	t->rval = render->rlayout & RL_CENTRE;
	render->rlayout |= RL_CENTRE;
}

static void centre_end(struct token_stack *t)
{
	render_newline_p();
	render->rlayout &= ~RL_CENTRE;
	render->rlayout |= t->rval;
}

static void dd_begin(struct token_stack *t)
{
	token_peek_pop("DT", "DL");
	render_newline_p();
}

/*
 *	In theory dir is not menu is not  ul. In practice dir is ul. 
 *	<dir> permits multiple columns but nobody does that either
 */
 
static void dir_begin(struct token_stack *t)
{
	render_newline_p();
	t->rval = render->rlist;
	render->rlist = RL_UL;
	render_indent();
}

static void dir_end(struct token_stack *t)
{
	render_unindent();
	render->rlist = t->rval;
	render_newline_p();
}

static void div_begin(struct token_stack *t)
{
	char *k = tag_attr_default(t, "align", "left");
	
	t->rval = render->rlayout & (RL_RIGHT|RL_CENTRE);
	render->rlayout &= ~RL_RIGHT|RL_CENTRE;
	if(strcasecmp(k, "right")==0)
		render->rlayout|=RL_RIGHT;
	if(strcasecmp(k, "center")==0)
		render->rlayout|=RL_CENTRE;
}

static void div_end(struct token_stack *t)
{
	render_newline_p();
	render->rlayout &= ~RL_RIGHT|RL_CENTRE;
	render->rlayout |= t->rval;
}

static void dl_begin(struct token_stack *t)
{
	render_newline_p();
	t->rval = render->rlist;
	render->rlist = RL_DL;
	render_indent();
}

static void dl_end(struct token_stack *t)
{
	render_unindent();
	render->rlist = t->rval;
	render_newline_p();
}
	
static void dt_begin(struct token_stack *t)
{
	token_peek_pop("DD", "DL");
	render_newline_p();
	b_begin(t);
}

static void dt_end(struct token_stack *t)
{
	b_end(t);
}

static void frame_begin(struct token_stack *t)
{
	int old_rfont = render->rfont;
	static char framestr[]="Frame: ";
	char *name = tag_attr_default(t, "name", "unnamed frame");
	char *url = tag_attr(t, "src");
	token_peek_pop("FRAME", "FRAMESET");
	
	if(url)
	{
		link_area_begin(t, L_URLGOTO, url);
		render->rfont |= RF_LINK;
		render->rfont &= ~RF_HIDE;
		render_text(framestr);
		render_text(name);
		render->rfont = old_rfont;
		link_area_end(t, L_URLGOTO);
		render_newline_p();
	}
}

static void frameset_begin(struct token_stack *t)
{
	render_newline_p();
	t->rval = render->rfont & RF_HIDE;
	render->rfont |= RF_HIDE;
}

static void frameset_end(struct token_stack *t)
{
	render_newline_p();
	render->rfont &= ~RF_HIDE;
	render->rfont |= t->rval;
}

static void head_begin(struct token_stack *t)
{
	render->rhead=1;
}

static void head_end(struct token_stack *t)
{
	render->rhead=0;
}


static void h_begin(struct token_stack *t, int n)
{
	char *k = tag_attr_default(t, "align", "left");
	
	t->rval = (render->rlayout & (RL_RIGHT|RL_CENTRE))|(render->rsize<<16);
	render->rlayout &= ~RL_RIGHT|RL_CENTRE;
	if(strcasecmp(k, "right")==0)
		render->rlayout|=RL_RIGHT;
	if(strcasecmp(k, "center")==0)
		render->rlayout|=RL_CENTRE;
	render->rsize = n;
	render_newline_p();
}

static void h_end(struct token_stack *t)
{
	render->rsize = t->rval >> 16;
	render->rlayout &= ~RL_CENTRE|RL_RIGHT;
	render->rlayout |= t->rval&0xFFFF;
	render_newline_p();
}

static void h1_begin(struct token_stack *t)
{
	h_begin(t,1);
}

static void h2_begin(struct token_stack *t)
{
	h_begin(t,2);
}

static void h3_begin(struct token_stack *t)
{
	h_begin(t,3);
}

static void h4_begin(struct token_stack *t)
{
	h_begin(t,4);
}

static void h5_begin(struct token_stack *t)
{
	h_begin(t,5);
}

static void h6_begin(struct token_stack *t)
{
	h_begin(t,6);
}

static void hide_begin(struct token_stack *t)
{
	t->rval = render->rfont & RF_HIDE;
	render->rfont |= RF_HIDE;
}

static void hide_end(struct token_stack *t)
{
	render->rfont &= ~RF_HIDE;
	render->rfont |= t->rval;
}

static void hr_begin(struct token_stack *t)
{
	static char linebuf[]="--------";
	
	centre_begin(t);
	/* Hack for now */
	render_text(linebuf);
	centre_end(t);
}

static void img_begin(struct token_stack *t)
{
	static char img[]="IMG ";
	static char space[]=" ";
	
	char *alt = tag_attr(t, "alt");

	if(alt)
	{
		render_text(alt);
		render_text(space);
	}
	else
		render_text(img);
}

static void li_begin(struct token_stack *t)
{
	static char o_box[]="O ";
	char buf[16];
	render_newline_p();
	if(render->rlist != RL_OL)
		render_text(o_box);
	else
	{
		snprintf(buf, 16, "%d ", render->rcount);
		render_text(buf);
	}
}

/*
 *	In theory dir is not menu is not  ul. In practice menu is ul. 
 */
 
static void menu_begin(struct token_stack *t)
{
	render_newline_p();
	t->rval = render->rlist;
	render->rlist = RL_UL;
	render_indent();
}

static void menu_end(struct token_stack *t)
{
	render_unindent();
	render->rlist = t->rval;
	render_newline_p();
}

static void noframe_begin(struct token_stack *t)
{
	t->rval = render->rfont & RF_HIDE;
	render->rfont &= ~RF_HIDE;
}

static void noframe_end(struct token_stack *t)
{
	render->rfont &= ~RF_HIDE;
	render->rfont |= t->rval;
}

static void ol_begin(struct token_stack *t)
{
	render_newline_p();
	t->rval = render->rlist | (render->rcount << 16);
	render->rcount = 1;
	render->rlist = RL_OL;
	render_indent();
}

static void ol_end(struct token_stack *t)
{
	render_unindent();
	render->rlist = (t->rval&0xFFFF);
	render->rcount = t->rval>>16;
	render_newline_p();
}

/*
 *	Paragraphs. The </P> is implied by the next <P>
 */
 
static void p_begin(struct token_stack *t)
{
	char *k = tag_attr_default(t, "align", "left");
	
	token_peek_pop(t->tag, NULL);
	
	t->rval = render->rlayout & (RL_RIGHT|RL_CENTRE);
	render->rlayout &= ~RL_RIGHT|RL_CENTRE;
	if(strcasecmp(k, "right")==0)
		render->rlayout|=RL_RIGHT;
	if(strcasecmp(k, "center")==0)
		render->rlayout|=RL_CENTRE;
	render_newline_p();
}

static void p_end(struct token_stack *t)
{
	render_newline_p();
	render->rlayout &= ~RL_RIGHT|RL_CENTRE;
	render->rlayout |= t->rval;
}

static void plain_begin(struct token_stack *t)
{
	t->rval = render->rfont & RF_PLAIN;
	render->rfont |= RF_PLAIN;
}

static void plain_end(struct token_stack *t)
{
	render->rfont &= ~RF_PLAIN;
	render->rfont |= t->rval;
}

static void pre_begin(struct token_stack *t)
{
	t->rval = render->rlayout & RL_PRE;
	render->rlayout |= RL_PRE;
}

static void pre_end(struct token_stack *t)
{
	render->rlayout &= ~RL_PRE;
	render->rlayout |= t->rval;
}

static void td_begin(struct token_stack *t)
{
	token_peek_pop("TD", "TR");
}

static void td_end(struct token_stack *t)
{
	char space[]=" ";
	render_text(space);
}


static void title_begin(struct token_stack *t)
{
	t->rval = render->rlayout & RL_TITLE;
	render->rlayout |= RL_TITLE;
}

static void title_end(struct token_stack *t)
{
	render->rlayout &= ~RL_TITLE;
	render->rlayout |= t->rval;
}

static void tr_begin(struct token_stack *t)
{
	token_peek_pop("TR", "TABLE");
}

static void tr_end(struct token_stack *t)
{
	render_newline_p();
}

static void ul_begin(struct token_stack *t)
{
	render_newline_p();
	t->rval = render->rlist;
	render->rlist = RL_UL;
	render_indent();
}

static void ul_end(struct token_stack *t)
{
	render_unindent();
	render->rlist = t->rval;
	render_newline_p();
}

static struct tag_render renderers[]={
	{"<DOCUMENT>", doc_begin, doc_end},
	{"!DOCTYPE", no_op, no_op},
	{"A", a_begin, a_end},
	{"B", b_begin, b_end},			/* Extension */
	{"BASE", base_begin, no_op},
	{"BLOCKQUOTE", bq_begin, bq_end},
	{"BODY", body_begin, body_end},
	{"BR", br_begin, no_op},
	{"CENTER", centre_begin, centre_end},
	{"DD", dd_begin, no_op},
	{"DIR", dir_begin, dir_end},
	{"DIV", div_begin, div_end},
	{"DL", dl_begin, dl_end},
	{"DT", dt_begin, dt_end},
	{"FORM", no_op, no_op},			/* Need to do */
	{"FRAME", frame_begin, no_op},		/* Frame hack */
	{"FRAMESET", frameset_begin, frameset_end},/* Frame hack */
	{"HEAD", head_begin, head_end},
	{"HTML", no_op, no_op},
	{"H1", h1_begin, h_end},
	{"H2", h2_begin, h_end},
	{"H3", h3_begin, h_end},
	{"H4", h4_begin, h_end},
	{"H5", h5_begin, h_end},
	{"H6", h6_begin, h_end},
	{"HR", hr_begin, no_op},
	{"IMG", img_begin, no_op},
	{"INPUT", no_op, no_op},		/* Need to do */
	{"LI", li_begin, no_op},
	{"MENU", menu_begin, menu_end},
	{"NOFRAMES", noframe_begin, noframe_end},	/* Frame hack */
	{"OL", ol_begin, ol_end},
	{"OPTION", no_op, no_op},		/* Need to do */
	{"P", p_begin, p_end},
	{"PLAINTEXT", plain_begin, plain_end},
	{"PRE", pre_begin, pre_end},
	{"SCRIPT", hide_begin, hide_end},	/* Extension */
	{"SELECT", no_op, no_op},		/* For now */
	{"STRONG", b_begin, b_end},		/* Extension */
	{"STYLE", hide_begin, hide_end},	/* This is needed because people like Oracle 
						   put style sheets in body areas which is not allowed . */
	{"TEXTAREA", no_op, no_op},		/* Need to do */
	{"TITLE", title_begin, title_end},
	{"TD", td_begin, td_end},		/* Table hack */
	{"TR", tr_begin, tr_end},		/* Table hack */
	{"UL", ul_begin, ul_end},
	{NULL, NULL, NULL}
};


struct tag_render *find_renderer(char *tag)
{
	struct tag_render *t = &renderers[0];
	while(t->tag)
	{
		if(strcasecmp(t->tag, tag)==0)
			return t;
		t++;
	}
	return NULL;
}
